from collections import Counter
from _dynamic import EndGapFree

def hashSeq(seq,kmer=12):
    '''
    Build a hash table of the kmer constituing the sequence seq
    :param seq: a dna sequence, supposed to be in uppercase
    :type seq: bytes
    :param kmer: the size of the considered kmer
    :type kmer: int
    :return: an associative array containing the hash table with kmer as key and an array
             of integer positions as value
    :rtype: a dict instance indexed by bytes and containing list of int
    '''
    lseq = len(seq)
    h = {}
    for i in range(lseq-kmer+1):
        w = seq[i:(i+kmer)]
        p = h.get(w,[])
        p.append(i)
        h[w]=p
    return h

def cmpHash(h1,h2,delta=2):
    '''
    Compare two hash tables as build by the hashSeq function.

    :param h1: the first hash table
    :type h1: a dict instance indexed by bytes and containing list of int
    :param h2: the second hash table
    :type h2: a dict instance indexed by bytes and containing list of int
    :param delta: delta around the main diagonal considered to compute the score
    :type delta: int
    :return: the shift corresponding to the main diagonal and the corresponding score
    :rtype: a tuple of two int 
    '''
    k1 = set(h1.keys())
    k2 = set(h2.keys())
    kc = k1 & k2
    stat = Counter()
    for k in kc:
        p1 = h1[k]
        p2 = h2[k]
        stat.update(i - j for i in p1 for j in p2)
    mstat = stat.most_common()
    if mstat:
        shift = mstat[0][0]
        
        score = 0
        for i in xrange(-delta+shift,shift+delta+1):
            score+=stat.get(i,0)
        mstat=(shift,score)
    else:
        mstat =(0,0)
    return mstat

def alignSequence(seq1,seq2,hash1=None,hash2=None,kmer=12,delta=2,smin=10,align=None):
    
    if align is None:
        align = EndGapFree()
    if hash1 is None:
        hash1=hashSeq(seq1, kmer)
    if hash2 is None:
        hash2=hashSeq(seq2, kmer)
        
    shift,score = cmpHash(hash1, hash2, delta)
    
    if score > smin:
        lseq1=len(seq1)
        lseq2=len(seq2)
        if shift < 0:
            direction=True
            subseq1 = seq1[0:shift]
            subseq2 = seq2[shift:]
            shift=-shift
            if subseq1==subseq2:
                ali = ([[0,shift],[lseq1,0]],[[lseq2,shift]])
            else:
                align.seqA=seq1
                align.seqB=seq2
                ali = align()
                shift = ali[0][0][1]
        else:
            direction=False
            subseq1 = seq1[shift:]
            subseq2 = seq2[0:lseq2-shift]
            if subseq1==subseq2:
                ali = ([[0,shift],[lseq2,0]],[[lseq1,shift]])
            else:
                align.seqA=seq2
                align.seqB=seq1
                ali = align()
                
#         if ali[0][0][0]==0:
#                 shift = ali[0][0][1]
#                 ali1  = (shift,ali[0][1:])
#         else:
#                 ali1  = (0,ali[0]) 
#                 
#         if ali[1][0][0]==0:
#                 shift = ali[1][0][1]
#                 ali2  = (shift,ali[1][1:])
#         else:
#                 ali2  = (0,ali[1]) 
                
        ali[0][-1][1]=0
        ali[1][-1][1]=0
        
        p=0
        gaps1={0:0}
        for n,g in ali[0]:
            p+=n
            if g > 0:
                gaps1[p]=g 
        p=0
        gaps2={0:0}
        for n,g in ali[1]:
            p+=n
            if g > 0:
                gaps2[p]=g 
            
    else:
        direction=True
        shift=0
        score=0
        gaps1=None
        gaps2=None

    return direction,score,gaps1,gaps2
    
def alignReads(readids,index,kmer=12,smin=10,delta=2):
    align = EndGapFree()

    rsize = index.getReadSize()
    nseq   = len(readids)
    seqs   = [index.getRead(j,0,rsize) for j in readids]
    hashes = [hashSeq(s,kmer) for s in seqs]
    
    scores = [(x[2][1],
               (readids[x[0]] if x[2][0] else readids[x[1]],x[2][2]),
               (readids[x[1]] if x[2][0] else readids[x[0]],x[2][3]),
              )
              for x in 
                  ((h1,h2,alignSequence(seqs[h1],seqs[h2],
                                hashes[h1],hashes[h2], 
                                kmer, delta, smin, align))
                  for h1 in xrange(nseq) 
                  for h2 in xrange(nseq) if h1 < h2)
              if x[2][1] >= smin]
    
    scores.sort(key=lambda x:x[0],reverse=True)
    
    return scores



def newGaps(newali,oldali):
    delta={}
    for p in newali:
        d = newali[p]-oldali.get(p,0)
        if d > 0:
            delta[p]=d 
    return delta
    
def readPos2Alignment(ali,read,pos):
    ref  = ali[read]
    apos = sum(x[1] for x in ref.iteritems() if x[0] <= pos) + pos
    return apos
    
def alignPos2Read(ali,read,pos):
    ref  = ali[read]
    posgap = ref.items()
    posgap.sort(key=lambda x:x[0])        
    
    b = sum(posgap[0])
    bornes = [(b,0)]
    
    for i,j in posgap:
        if i > 0:
            b+=i 
            bornes.append((b,bornes[-1][1]))
            b+=j
            bornes.append((b,i))

#    print bornes
    
    if pos < bornes[0][0]:
#        tag = "*"
        rpos=0
    else:
        i=0
        lb = len(bornes)
        while i < lb and pos > bornes[i][0]:
            i+=1
            
        if i == lb:
#            tag = "@"
            rpos = pos - bornes[-1][0] + bornes[-1][1]
        else:
            isnuc = i & 1
            if isnuc:
#                tag = "+"
                rpos = pos - bornes[i-1][0] + bornes[i][1]
            else:
#                tag = "-"
                rpos = bornes[i][1] 
            
    return rpos
    
def applyGaps(key,gaps,ali,rsize):
    
    
    if gaps:
#         time.sleep(1)
#         print key,gaps,ali,rsize
        for pos,length in gaps.iteritems():
            pali = readPos2Alignment(ali, key, pos)
            for read in ali:
                pread = alignPos2Read(ali, read, pali)
                if pread < rsize:
                    ali[read][pread]=ali[read].get(pread,0)+length
#        print ">>>",ali
    return ali
                    
def mergeAlignments(common, ali1,ali2,rsize):
    r1 = ali1[common]
    r2 = ali2[common]
    
    lr1 = len(r1)
    lr2 = len(r2)
    
    if lr1 > lr2:
        r1,r2 = r2,r1
        ali1,ali2 = ali2,ali1
        
    delta = r2[0] - r1[0]
        
    for r in ali1.itervalues():
        r[0]+=delta 
        
    ngap1 = newGaps(r1, r2)
    ngap2 = newGaps(r2, r1)
    

    applyGaps(common,ngap1,ali2,rsize)
    applyGaps(common,ngap2,ali1,rsize)
    
#     if common==-15208645:
#         print "+>>>",ali1,ali2
#         print '\n'.join(alignment2bytes(ali1,r))
#         print
#         print '\n'.join(alignment2bytes(ali2,r))

    ali2.update(ali1)
    
#     if common==-15208645:
#         print "@>>>",ali2
#         print '\n'.join(alignment2bytes(ali2,r))
#         time.sleep(5)

    return ali2
    
    
    
    

def multiAlignReads(readids,index,kmer=12,smin=10,delta=2):
    readids = list(set(readids))
    scores=alignReads(readids,index,kmer,smin,delta)
    rsize = index.getReadSize()
    
    alignment={}
    for (score,(r1,ali1),(r2,ali2)) in scores:
        partial1 = [x for x in alignment.items() if r1 in x[0]]
        partial2 = [x for x in alignment.items() if r2 in x[0]]

        if not partial1 or (partial1 and not r2 in partial1[0][0]):
            # the two reads are not already in the same alignment
            if partial1:
                del alignment[partial1[0][0]]
                partial1 = partial1[0][1]
                tmp      = {r1:ali1, r2:ali2}
    #            print r1, partial1, tmp
                partial1 = mergeAlignments(r1, partial1, tmp, rsize)
            else:
                partial1 = {r1:ali1, r2:ali2}
                
    
            if partial2:
                del alignment[partial2[0][0]]
                partial2 = partial2[0][1]
                merged   = mergeAlignments(r2, partial1, partial2, rsize)
            else:
                merged   = partial1
                
            alignment[frozenset(merged)]=merged
        
    for k in alignment:
        al = alignment[k]
        minshift = min(i[0] for i in al.values())
        for i in al:
            al[i][0]=al[i][0]-minshift
    
    return alignment.values()

def alignment2bytes(ali,index):
    i = 0
    rsize = index.getReadSize()
    res = []
    for r in ali:
        seq = index.getRead(r,0,rsize)
        gaps = ali[r].items()
        gaps.sort(key=lambda g:g[0])
        aseql=[]
        for pos,length in gaps:
            if pos>0:
                aseql.append(seq[i:pos])
                aseql.append('-'*length)
            else:
                aseql.append('+'*length)                
            i=pos
        aseql.append(seq[i:rsize])
#        print ali[r],''.join(aseql)
        res.append(''.join(aseql))
        res.sort()      

    return res


def consensus(aln,index,covmin=5):
    res = alignment2bytes(aln,index)
    lali = max(len(i) for i in res)
    seq = [Counter(l for l in (s[pos:pos+1] for s in res) if l and l in '-ACGT').most_common()[0] for pos in xrange(lali)]
    seq = [i for i in seq if i[0]!='-']
    coverage = [i[1] for i in seq]
    seq = ''.join([i[0] for i in seq])
    i=0
    while coverage[i] < covmin:
        i+=1
    j=len(coverage)-1
    while coverage[j] < covmin:
        j-=1
        
    if i < j:
        seq = seq[i:j+1]
    return seq        
     
