"""
.. module:: orgasm.tango
   :platform: Unix
   :synopsis: The :py:mod:`~orgasm.tango` package contains a set of functions useful to manage assembling structure.

The :py:mod:`orgasm.tango` package
==================================

The :py:mod:`~orgasm.tango` package contains a set of functions useful to manage assembling structure.

.. moduleauthor:: Eric Coissac <eric.coissac@inria.fr>

:author:  Eric Coissac
:contact: eric.coissac@inria.fr


.. warning::

    The :py:mod:`~orgasm.tango` package functions aims to be integrated 
    into other packages as standalone functions of class methods.
"""
import sys
from collections import Counter
from itertools import imap

def tango(self,seeds,
          minread=20,    minratio=0.1, 
          mincov=2,      minoverlap=40,
          lowfilter=True,maxjump=1,restrict=None):
    '''
    
    :param seeds: a list describing the extensions points
    :type seeds: a dict indexed by +readid of tuples composed of :
                    - the read sequence
                    - the rev.comp read sequence
                    - gene name
                 the two first elements can be set to None.
                 In this case sequences have to be extracted
                 from the index
                 if the last element is None then no gene is associated
                 to this read
    :param minread:
    :type minread:
    :param minratio:
    :type minratio:
    :param mincov:
    :type mincov:
    :param minoverlap:
    :type minoverlap:
    :param lowfilter:
    :type lowfilter:
    :param maxjump:
    :type maxjump:    
    :param restrict:
    :type restrict:    
    '''
        
    # a direct pointer to the read index
    index = self.index
    
    # a direct pointer to the graph structure
    graph = self.graph
    
    # a direct access to the read length
    readLength = index.getReadSize()
    readmax    = len(index) + 1
    
    # extension cycle counter for the progress status monitoring
    cycle=0
    
    # count of missing reads in the graph (aka fake reads)
    fake=0
    
    componantes=0
    
    # list of the stack size during the last 1000 cycles 
    lastcycle=[]
    
    # count of pair end jump on an extension branch
    maxanchor = 0
    
    while(seeds):
        
        #time.sleep(1)
        cycle+=1
        
        # Counter used to check how many seed will be added 
        # to the stack during the cycle
        addedseed=0

        # We pop one seed from the stack
        
        readid,(gene,) = seeds.pop()
        
        # We test if the read poped out from the stack
        # is just a jump mark
        if readid==0:
            maxanchor-=1
            continue
        
        if readid < 0 or readid not in graph:
            
            if readid < 0:
                readid=-readid
            #print "\n==>",readid,str(gene)
            #
            # We add the read into the graph
            #
            
            lastcycle.append(len(seeds))
            if len(lastcycle) > 1000:
                lastcycle=lastcycle[-1000:]         
            print >>sys.stderr,b"Cycle : %8d  (%d nodes / %4.1f%% fake) Waiting points : %8d / %6.2f  Gene: %s \r" % (cycle,len(graph),float(fake)/(len(graph)+1)*200.,len(seeds),float(sum(lastcycle))/len(lastcycle),gene),
            #print >>sys.stderr,b"Cycle : %8d  (%d nodes) Waiting points : %8d    " % (cycle,len(graph),len(seeds))
    
                
            #
            # extends the read on the right side
            #
                
            rextensions = index.checkedExtensions(index.getRead(readid,1,readLength-1), 
                                                 minread, minratio, 
                                                 mincov,  minoverlap, 
                                                 1,
                                                 restrict=restrict)
            
            #
            # Build a bijective relationship for the right extensions
            #
            
            for k in rextensions.keys():
                backe = index.checkedExtensions(index.getRead(-rextensions[k][1][0],1,readLength-1),
                                                minread, minratio, 
                                                mincov,  minoverlap, 
                                                1,
                                                restrict=restrict)
                back = set(reduce(lambda p,q:p+q,[backe[y][1] for y in backe],[]))
                
                if -readid not in back:
                    #print "Remove R extension : ",k
                    del rextensions[k]
            
    
            lextensions = index.checkedExtensions(index.getRead(-readid,1,readLength-1), 
                                                 minread, minratio, 
                                                 mincov,  minoverlap, 
                                                 1,
                                                 restrict=restrict)
     
            #
            # Build a bijective relationship for the left extensions
            #
    
            for k in lextensions.keys():
                backe = index.checkedExtensions(index.getRead(-lextensions[k][1][0],1,readLength-1), 
                                                 minread, minratio, 
                                                 mincov,  minoverlap, 
                                                 1,
                                                 restrict=restrict)
                
                back = set(reduce(lambda p,q:p+q,[backe[y][1] for y in backe],[]))
                
                if readid not in back:
                    #print "Remove L extension : ",k
                    del lextensions[k]
                    
    #        print "Rext = ",rextensions
    #        print "Lext = ",lextensions
    
            if rextensions or lextensions:
                node  = graph.addNode(readid)
                node['gene']=gene
                if readid < readmax:
                    node['fake5']=0
                    node['fake3']=0
                else:
                    fake+=1
                    node['graphics']={'fill':"#00FF00"}
    
    #        print "R",len(seq)
    
            for ex in rextensions:
                data = rextensions[ex]
                coverage = data[0]
    #            print ex,data
                
                nodeindexE,c,s = index.getIds(data[1][0])
                                    
                if nodeindexE in graph:
                    if data[1][0] not in s:
                        nodeindexE=-nodeindexE
                    if readid!=-nodeindexE:
                        edges = graph.addEdge(readid,nodeindexE)
                        edges[1][b'coverage']=coverage
                        edges[2][b'coverage']=coverage
                        edges[1][b'label']="%s (%d)" % (edges[1][b'ext'],coverage)
                        edges[2][b'label']="%s (%d)" % (edges[2][b'ext'],coverage)
                        edges[1][b'type']=b"next read"
                        edges[2][b'type']=b"next read"
                    else:
                        print >>sys.stderr,"\nWARNING : self loop on %d" % nodeindexE
      #              print "linked"
                else:
      #              print "not linked"
                    seeds.append((nodeindexE,(None,)))
                    addedseed+=1
                                
            #
            # extends the read on the left side
            # which is equivalent to extend on the right side
            # the reverse complement of the read
            #
            
      #      print "L",len(cseq)
            
     
            for ex in lextensions:
                data = lextensions[ex]
                coverage = data[0]
    #            print ex,data
    
                nodeindexE,c,s = index.getIds(data[1][0])
                                        
                if nodeindexE in graph:
                    if data[1][0] not in s:
                        nodeindexE=-nodeindexE
                    edges = graph.addEdge(-readid,nodeindexE)
                    edges[1][b'coverage']=coverage
                    edges[2][b'coverage']=coverage
                    edges[1][b'label']="%s (%d)" % (edges[1][b'ext'],coverage)
                    edges[2][b'label']="%s (%d)" % (edges[2][b'ext'],coverage)
                    edges[1][b'type']=b"next read"
                    edges[2][b'type']=b"next read"
      #              print "linked"
                else:
      #              print "not linked"
                    seeds.append((nodeindexE,(None,)))
                    addedseed+=1
    
            if addedseed==0 and maxjump > 0 and restrict is None:
                if maxanchor < maxjump and readid < readmax and readid in graph:
                    maxanchor+=1
                    seeds.append((0,(None,)))
                    anchors = map(lambda id:index.getIds(index.getPairedRead(id))[0],index.getIds(readid)[2])
                    
                    print >>sys.stderr,"\n\nJumpGap on read %d\n" % readid
                    for a in anchors:                        
                        if a not in graph:
                            seeds.append((-a,(None,)))
                        
    return graph

def cutLowCoverage(self,mincov,terminal=True):
    def isTerminal(g,n):
        return len(list(g.parentIterator(n)))==0 or len(list(g.neighbourIterator(n)))==0
    
    if terminal:
        tstates=[True]
    else:
        tstates=[True,False]
    cg = self.compactAssembling()
    index = self.index
    readSize = index.getReadSize()
    extremities = self.endNodeSet() | self.startNodeSet()
    for terminal in tstates:
        go = True
        while go:
            go = False
            stems = [x for x in cg.edgeIterator() 
                     if not terminal or (isTerminal(cg, x[0]) or isTerminal(cg, x[1]))]
            if stems:
                stems.sort(key=lambda i:cg.getEdgeAttr(*i)['weight'],reverse=True)
                lightest = stems.pop()
                lattr = cg.getEdgeAttr(*lightest)
        #         print >>sys.stderr,lightest[0],lightest[1]
        #         print >>sys.stderr,lattr['weight'] < mincov
        #         print >>sys.stderr,isTerminal(cg, lightest[0]), isTerminal(cg, lightest[1])
        #         print >>sys.stderr,lattr['weight'] < mincov and ((not terminal) or isTerminal(cg, lightest[0]) or isTerminal(cg, lightest[1]))
        #         print >>sys.stderr
                if lattr['weight'] < mincov: 
                    if stems:
                        go=True
                    if lattr['stemid'] > 0:
                        for n in lattr['path'][1:-1]:
                            if n in self.graph:
                                del self.graph[n]
                        if lightest[0] in extremities and lightest[0] in self.graph:
                            del self.graph[lightest[0]]
                        if lightest[1] in extremities and lightest[1] in self.graph:
                            del self.graph[lightest[1]]
                        
                    print >>sys.stderr,"Remaining edges : %d node : %d\r" % (self.graph.edgeCount(),len(self)),
        
                    cg.deleteEdge(lightest[0],lightest[1],lightest[2])
                    tojoin=[]
                    if lightest[0] in extremities:
                        del cg[lightest[0]]
                    else:
                        tojoin.append(lightest[0])
                    if lightest[1] in extremities:
                        del cg[lightest[1]]
                    else:
                        tojoin.append(lightest[1])
                    for c in tojoin:
                        begin = list(cg.parentIterator(c))
                        end   = list(cg.neighbourIterator(c))
                        if len(begin)==1 and len(end)==1:
                            attr1 = cg.getEdgeAttr(begin[0],c)
                            attr2 = cg.getEdgeAttr(c,end[0])
                            lentot= attr1['length'] + attr2['length']
                            weight= (attr1['length'] * attr1['length'] + attr2['length'] * attr2['length'])/lentot
                            sequence=attr1['sequence'] + attr2['sequence']
                            stemid=lattr['stemid']
                            path=attr1['path'] + attr2['path'][1:]
                            graphics={b'width':int(weight/readSize)+1,
                                      b'arrow':b'last'}
                            
                            if lentot > 10:
                                label=b"%d : %s->(%d)->%s  [%d]" % (stemid,sequence[0:5],lentot,sequence[-5:],weight)
                            else:
                                label=b"%d : %s->(%d)  [%d]" % (stemid,sequence,lentot,weight)
                                
                            del cg[c]
                            
                            cg.addEdge(begin[0],end[0])
                            attr = cg.getEdgeAttr(begin[0],end[0])
                            attr[b'sequence']=sequence
                            attr[b'stemid']=stemid
                            attr[b'path']=path
                            attr[b'length']=lentot
                            attr[b'weight']= weight
                            attr[b'graphics']=graphics
                            attr[b'label']=label
        
            
  
def matchtoseed(matches,index):
    k = matches.keys()
    s=[]
    for x in k:
        s.extend((index.getIds(y[0])[0],(x,)) for y in matches[x])  

    return s

def scaffold(self,assgraph,minlink=5,back=200,addConnectedLink=False):
    '''
    Add relationships between edges of the assembling graph related to the
    par ended links.
    
    :param assgraph: The compact assembling graph as produced by the
                     :py:meth:`~orgasm.assembler.Assembler.compactAssembling` method
    :type assgraph:  :py:class:`~orgasm.graph.DiGraphMultiEdge`
    :param minlink:  the minimum count of pair ended link to consider 
                     for asserting the relationship
    :type minlink:   :py:class:`int`
    :param back:     How many base pairs must be considered at the end of each edge
    :type back:      :py:class:`int`
    :param addConnectedLink: add to the assembling graph green edges for each directly 
                             connected edge pair representing the pair ended links 
                             asserting the connection.
    :type addConnectedLink: :py:class:`bool`
    '''
#    o = [i for i in assgraph.edgeIterator(edgePredicate=lambda e: 'stemid' not in assgraph.getEdgeAttr(*e))]   
#     for i in o:
#         assgraph.deleteEdge(i[0],i[1],i[2])
        
    e = [i for i in assgraph.edgeIterator(edgePredicate=lambda e: 'stemid' in assgraph.getEdgeAttr(*e))]
    eid = [assgraph.getEdgeAttr(*i)['stemid'] for i in e]
    ne = len(e)
    s=[]
    for e1 in xrange(ne):
        for e2 in xrange(ne):
            connected = e[e1][1]==e[e2][0]
            linkedby  = pairEndedConnected(self,assgraph, eid[e1], eid[e2], back)
            
            if connected and addConnectedLink:
                if linkedby > minlink:
                    s.append((e[e1][1],e[e2][0],linkedby,eid[e1], eid[e2],"#00FF00"))
                    print >>sys.stderr,"%d -> %d connection asserted by %d pair ended links" % (eid[e1], eid[e2],linkedby)
                else:
                    print >>sys.stderr,"%d -> %d connection not asserted by pair ended link" % (eid[e1], eid[e2])
                    
            if not connected and linkedby > minlink:
                s.append((e[e1][1],e[e2][0],linkedby,eid[e1], eid[e2],"#FF0000"))
                print >>sys.stderr,"%d -> %d scaffolded by %d pair ended links" % (eid[e1], eid[e2],linkedby)
    
    for x,y,z,s1,s2,color in s:
        attr = assgraph.addEdge(x,y)
        attr[b'label']=b"Gap (%d)  %d -> %d" % (z,s1,s2)
        attr[b'graphics']={b'width':z / 10.,
                           b'arrow':b'last',
                           b'fill':color
                          }
 
def insertFragment(self,seq):
    index = self.index
    rsize = index.getReadSize()
    readmax=len(index)+1
    seeds = set()
    
    if len(seq) >= rsize:
        graph = self.graph
        probe = seq[0:rsize]
        readid = index.getReadIds(probe)
        
        for i in xrange(1,len(seq)-rsize+1):
            coverage = readid[1]
            ireadid   = readid[0]
            seeds.add(ireadid)
            
            if ireadid not in graph:
                node  = graph.addNode(ireadid)
                if ireadid < readmax:
                    node['fake5']=0
                    node['fake3']=0
                else:
                    node['graphics']={'fill':"#0000FF"}
                
            probe = seq[i:i+rsize]
            readidE = index.getReadIds(probe)
            coverage = readidE[1]
            ireadidE   = readidE[0]
            
            if ireadidE not in graph:
                node  = graph.addNode(ireadidE)
                if ireadidE < readmax:
                    node['fake5']=0
                    node['fake3']=0
                else:
                    node['graphics']={'fill':"#0000FF"}
                    
            
            if ireadid!=-ireadidE:
                edges = graph.addEdge(ireadid,ireadidE)
                edges[1][b'coverage']=coverage
                edges[2][b'coverage']=coverage
                edges[1][b'label']="%s (%d)" % (edges[1][b'ext'],coverage)
                edges[2][b'label']="%s (%d)" % (edges[2][b'ext'],coverage)
                edges[1][b'type']=b"next read"
                edges[2][b'type']=b"next read"
            else:
                print >>sys.stderr,"\nWARNING : self loop on %d" % readidE
        
            readid = readidE
            
        seeds.add(ireadidE)
        
    return [(i,("Consensus",)) for i in seeds]
                
def extendStem(self,stem,cg,back,kmer):
    def lowcomplexity(x):
        return x[1:]==x[0:-1] or x[2:]==x[0:-2] or x[3:]==x[0:-3]
    r = self.index
    rsize = r.getReadSize()
    rcount =len(r)
    
    pa=cg.getEdgeAttr(stem[0], stem[1], stem[2])['path']
    sp = dict((j,r.getRead(j,0,rsize)) 
             for j in (r.getPairedRead(i) for i in pa[-back:] if abs(i) <= rcount) 
             if j not in asm.graph)
    c = Counter()
    for i in sp.values():    
        c.update(k for k in (i[j:j+kmer] for j in xrange(0,rsize-kmer)) if not lowcomplexity(k))

    s = []
    if c:
        m=c.most_common()[0]
        
        if m[1]>1:       
            s = [(abs(j),(None,)) 
                 for j in [i[0] for i in sp.items() if m[0] in i[1]]]

    return s

      
def pairendedJump(asm,back=300,kmer=20):
    
    cg = asm.compactAssembling()
    endFree = [e for e in cg.edgeIterator() if len(list(cg.neighbourIterator(e[1])))==0]
    s = []
    
    for e in endFree:
        s.extend(extendStem(asm,e,cg,back,kmer))
        
    return s


def getPairedRead(self,assgraph,stemid,back,end=True): 
    '''
    
    :param assgraph:
    :type assgraph:
    :param stemid:
    :type stemid:
    :param back:
    :type back:
    :param end:
    :type end:
    '''
    r = self.index 
    if not end:
        stemid=-stemid  
    stem = assgraph.edgeIterator(edgePredicate=lambda e:assgraph.getEdgeAttr(e[0],e[1],e[2])['stemid']==stemid).next()
    path=assgraph.getEdgeAttr(*stem)['path']
    path=path[-back:]
    reads = [(i,) + r.getIds(i) for i in path if i <= len(r)]
    reads=[i[3] if i[0] in i[3] else set(-j for j in i[3]) for i in reads]
    paired=[(r.getPairedRead(j) for j in i) for i in reads]
    paired=[[(j,) + r.getIds(j) for j in i if j!=0] for i in paired]
    paired = [[j[1] if j[0] in j[3] else -j[1] for j in i] for i in paired]
    paired = set(reduce(lambda a,b: a+b,(i for i in paired)))
    if not end:
        paired = set(-i for i in paired)
    return paired

def pairEndedConnected(self,assgraph,edge1,edge2,back=200):
    '''
    Returns how many pair ended reads link two edges in a compact assembling graph
    
    :param assgraph: The compact assembling graph as produced by the
                     :py:meth:`~orgasm.assembler.Assembler.compactAssembling` method
    :type assgraph:  :py:class:`~orgasm.graph.DiGraphMultiEdge`
    :param edge1:    The ``stemid`` of the first edge
    :type edge1:     :py:class:`int`
    :param edge2:    The ``stemid`` of the second edge
    :type edge2:     :py:class:`int`
    :param back:     How many base pairs must be considered at the end of each edge
    :type back:      :py:class:`int`
    
    :return:         The count of pair ended reads linking both the edges
    :rtype:          :py:class:`int`
    '''
    
    p1 = getPairedRead(self,assgraph,edge1,back,end=True)
    
    s2 = assgraph.edgeIterator(edgePredicate=lambda e:assgraph.getEdgeAttr(e[0],e[1],e[2])['stemid']==edge2).next()
    p2 = set(assgraph.getEdgeAttr(*s2)['path'][0:back])
    
    n = len(p1 & p2)
    
    return n
    
def coverageEstimate(self):
    def weight(a,b):
        maxw=0
        for e in cg.edgeIterator(edgePredicate = lambda i:i[0]==a and i[1]==b):
            attr = cg.getEdgeAttr(*e)
            w = attr['length'] * attr['weight']
            if w > maxw:
                maxw=w        
        return maxw 
    
    def coverage(start):
        wp = cg.minpath(start,distance=weight,allowCycle=True)
#        print ">>>",start,wp
        maxpath = max(wp[0].values())
        maxend=[i for i in wp[0] if wp[0][i]==maxpath][0]
        i=maxend
        path=[i]
        if i == start and wp[0][i] > 0:
            i=wp[1][i]
            path.append(i)
        while i!=start:
            i=wp[1][i]
            path.append(i)
        return(maxpath,path)
    
    cg = self.compactAssembling()
    paths = [coverage(i) for i in cg]
#    print paths
    maxpath=0
    path=None
    for w,p in paths:
        if w > maxpath:
            maxpath=w
            path=p 
#    print path        
    path.reverse()
    j = path[0]
    cumlength=0
    for i in path[1:]:
        maxw=0
        maxlength=0
        for e in cg.edgeIterator(edgePredicate = lambda k:k[0]==j and k[1]==i):
            attr = cg.getEdgeAttr(*e)
            w = attr['length'] * attr['weight']
#            print w , attr['length'] , attr['weight']
            if w > maxw:
                maxw=w
                maxlength=attr['length']       

        cumlength+=maxlength
        j=i
        
    return maxpath,cumlength,maxpath/float(cumlength)

def path2fasta(self,assgraph,path,identifier="contig",minlink=10,nlength=20,back=200):
    '''
    Convert a path in an compact assembling graph in a fasta formated sequences.
    
    :param assgraph: The compact assembling graph as produced by the
                     :py:meth:`~orgasm.assembler.Assembler.compactAssembling` method
    :type assgraph:  :py:class:`~orgasm.graph.DiGraphMultiEdge`
    :param path:     an ``iterable`` providing an ordered list of ``stemid`` indicating
                     the path to follow.  
    :type path:      an ``iterable`` over :py:class:`int`
    :param identifier: the identifier used in the header of the fasta formated sequence
    :type identifier:  :py:class:`bytes`
    :param minlink:  the minimum count of pair ended link to consider 
                     for asserting the relationship
    :type minlink:   :py:class:`int`
    :param nlength:  how many ``N`` must be added between two segment of sequences only connected
                     by pair ended links
    :type nlength:   :py:class:`int`
    :param back:     How many base pairs must be considered at the end of each edge
    :type back:      :py:class:`int`
    
    :returns: a string containing the fasta formated sequence
    :rtype: :py:class:`bytes`

    :raises: :py:class:`AssertionError`
    '''
    def getEdge(stemid):
        stem = assgraph.edgeIterator(edgePredicate=lambda e:assgraph.getEdgeAttr(e[0],e[1],e[2])['stemid']==stemid).next()
        return stem
        
        
    coverage=[]
    slength=[]
    seq=[]
    label=[]
    oldstem=None
    oldid=None
    
    for stemid in path:
        stem = getEdge(stemid)
        attr = assgraph.getEdgeAttr(*stem)

        if oldstem is not None:
            connected = pairEndedConnected(self,assgraph,oldid,stemid,back)
            
            if oldstem[1]==stem[0]:
                print >>sys.stderr,"Both segments %d and %d are connected" % (oldid,stemid)
                label.append('{connection: %d}' % connected)
            else:
                print >>sys.stderr,"Both segments %d and %d are disconnected" % (oldid,stemid)
                if connected > minlink:
                    print >>sys.stderr,"But linked by %d pair ended links" % connected
                    seq.append('N'*nlength)
                    label.append('{N-connection: %d}' % connected)
                else:
                    raise AssertionError('Disconnected path between stem '
                                         '%d and %d only %d pair ended links' % (oldid,stemid,connected))
        
        seq.append(attr['sequence'])
        coverage.append(attr['weight'])
        slength.append(attr['length'])
        label.append(attr['label'])
        
        oldstem = stem
        oldid=stemid
        
    s1 = getEdge(path[-1])
    s2 = getEdge(path[0])
    sid1=assgraph.getEdgeAttr(*s1)['stemid']
    sid2=assgraph.getEdgeAttr(*s2)['stemid']
    connected = pairEndedConnected(self,assgraph,sid1,sid2,back)
    
    if s1[1]==s2[0]:
        print >>sys.stderr,"Path is circular"
        circular=True
        label.append('{connection: %d}' % connected)
    else:
        if connected > minlink:
            print >>sys.stderr,"Path is circular but disconnected" 
            print >>sys.stderr,"Linked by %d pair ended links" % connected
            label.append('{N-connection: %d}' % connected)
            circular=True
        else:
            print >>sys.stderr,"Path is linear" 
            circular=False
   
        
    sequence = ''.join(seq)
    length = sum(slength)
    mcov = sum(a * b for a,b in imap(None,slength,coverage)) / float(length)
    
    fasta=[">%s seq_length=%d; coverage=%5.1f; circular=%s %s" % (identifier,length,
                                                                  mcov,bytes(circular),
                                                                  '.'.join(label))]
    fasta.extend(sequence[i:i+60] for i in xrange(0,len(sequence),60))
    return "\n".join(fasta)
    

